from typing import Optional, List, Any, Dict, Union, Tuple, Iterable, NamedTuple
import time
import random
import logging
from threading import Lock
from ftdi_serial import Serial, Device, SerialException, SerialReadTimeoutException
from north_utils.crc import build_c9_crc

logger = logging.getLogger(__name__)


C9_NUM_AXIS = 10
C9_FLAG_AXIS_MOVING = '1'
C9_AXIS_MOVING_SLEEP_TIME = 0.001


FlagsDict = Dict[str, Union[int, bytes, float]]


class C9ControllerError(Exception):
    pass


class C9Error(C9ControllerError):
    """
    Errors sent from the C9 Controller
    """

    SUCCESS = 0
    TIMEOUT = 1
    INVALID_COMMAND = 2
    INVALID_ARGUMENT = 3
    INVALID_SEPARATOR = 4
    TOO_MANY_ARGUMENTS = 5
    CMD_NOT_FOUND = 6
    INVALID_ARGUMENTS = 7
    INVALID_AXIS = 8
    INVALID_POSITION = 9
    INVALID_VELOCITY = 10
    INVALID_ACCELERATION = 11
    AXIS_MOVING = 12
    EEPROM_READ = 13
    EEPROM_WRITE = 14
    AXIS_FEATURE_UNAVAILABLE = 15
    MODBUS_TIMEOUT = 16
    MODBUS_INVALID_RESPONSE = 17
    INVALID_TEMP_CONT = 18
    INVALID_JOINT = 19
    INVALID_CRC = 20
    HOMING = 21
    SCALE_ERROR = 34

    CONTROLLER_DEADLOCK = 50

    def __init__(self, error_number, error_message, *args, **kwargs):
        super().__init__(error_message, *args, **kwargs)
        self.error_number = error_number


class C9ResponseError(C9ControllerError):
    pass


def dict_to_value_list(items: Dict[Any, Any]) -> List[Any]:
    """
    Convert a dictionary into a flattened value list containing keys and values in sequence

    :param items: dictionary to convert
    :return: list with keys and values from items dictionary
    """

    values = []
    for k, v in items.items():
        values += [k, v]

    return values


def retry_timeout_scale(total_retries: int, remaining: int, scale: float=0.3) -> float:
    """
    Calculate an exponential retry timeout based on the max and remaining number of retries

    :param total_retries: total number of retries allowed
    :param remaining: number of retries remaining
    :param scale: optional scale for exponential, defaults to 0.3
    :return: retry timeout in seconds
    """

    return 2 ** ((total_retries - remaining) * scale)


class C9ControllerStats:
    def __init__(self):
        self.requests: int = 0
        self.responses: int = 0
        self.commands: int = 0
        self.errors: int = 0
        self.error_types: Dict[int, int] = {}

    @property
    def error_types_str(self):
        return ', '.join([f'{k}={v}' for k, v in self.error_types.items()])

    def add_error(self, error: Exception):
        error_number = getattr(error, 'error_number', -1)
        self.error_types[error_number] = self.error_types.get(error_number, 0) + 1


class C9Controller:
    NUM_MAIN_AXES = 4
    BIAS_MIN_SHOULDER = 0
    BIAS_MAX_SHOULDER = 1
    BIAS_CLOSEST = 2

    AXIS_GRIPPER = 0
    AXIS_ELBOW = 1
    AXIS_SHOULDER = 2
    AXIS_COLUMN = 3
    AXIS_CAROUSEL = 4

    RETRY_ERRORS = [
        C9Error.INVALID_CRC,
        C9Error.INVALID_ARGUMENT,
        C9Error.INVALID_ARGUMENTS,
        C9Error.INVALID_SEPARATOR,
        C9Error.INVALID_COMMAND,
        C9Error.CMD_NOT_FOUND,
    ]

    default_controller = None
    locks: Dict[Device, Lock] = {}

    @classmethod
    def get_lock(cls, device: Device):
        if device in cls.locks:
            return cls.locks[device]

        cls.locks[device] = Lock()
        return cls.locks[device]

    def __init__(self,
                 connection: Serial=None,
                 address: Optional[int]=1,
                 device: Device=None,
                 device_serial: str=None,
                 device_number: int=None,
                 ping: bool=True,
                 home: bool=False,
                 connect: bool=True,
                 write_timeout: float=0.5,
                 read_timeout: float=0.5,
                 retry_timeout: float=0.5,
                 lock_timeout: float=120,
                 move_wait_delay: float=0.1,
                 move_wait_scale: float=0.8,
                 move_prediction: bool=True,
                 retries: int=10,
                 command_delay: float=0.0015,
                 debug_protocol: bool=False,
                 paused: bool=False,
                 use_joystick: bool=True,
                 ignore_missing_joystick: bool=True,
                 verbose=False) -> None:

        try:
            self.connection = connection if connection is not None else \
                Serial(device=device, device_serial=device_serial, device_number=device_number, connect=connect,
                       read_timeout=read_timeout, write_timeout=write_timeout)
        except SerialException as err:
            if self.connection and self.connection.device is not None:
                self.connection.device.close()
            raise err

        self.logger = logger.getChild(self.__class__.__name__)
        self.device_logger = logger.getChild(self.__class__.__name__ + '.device')
        self.network_address = address
        self.axis_handlers = {}
        self.output_handlers = {}
        self.verbose = verbose
        self.command_error_count = {}
        self.retries = retries
        self.read_timeout = read_timeout
        self.write_timeout = write_timeout
        self.retry_timeout = retry_timeout
        self.lock_timeout = lock_timeout
        self.move_wait_delay = move_wait_delay
        self.move_wait_scale = move_wait_scale
        self.move_prediction = move_prediction
        self.sequence_number = random.choice(range(1, 99))
        self.debug_protocol = debug_protocol
        self.paused = paused
        self.lock = self.get_lock(self.connection.device)
        self.command_delay = command_delay
        self.stats = C9ControllerStats()

        if connect:
            try:
                if ping:
                    self.ping()
            except C9Error:
                if not self.debug_protocol:
                    self.debug_protocol = True
                    self.ping()

        if C9Controller.default_controller is None:
            C9Controller.default_controller = self

        if use_joystick:
            from north_c9.joysticks import N9Joystick
            self.joystick = N9Joystick(self, ignore_missing_joystick=ignore_missing_joystick)

        if home:
            self.home()

    def connect(self):
        self.connection.connect()

    def disconnect(self):
        self.connection.disconnect()
        if self.joystick.running:
            self.joystick.stop()

    def is_aux_axis(self, axis):
        return axis >= self.NUM_MAIN_AXES

    def encode_command(self, name: str, flags_dict: FlagsDict, args: List[Any], sequence_number: int):
        args = [arg for arg in args if arg is not None]
        command = [name]
        command += self.build_flags(flags_dict)
        command += [str(round(arg)) for arg in args if arg is not None]
        command_str = ' '.join(command) + '\r'

        if not self.debug_protocol:
            command_str = str(sequence_number) + ' ' + command_str
            crc = build_c9_crc(command_str.encode()).decode()
            header_str = ' '.join([f'@{self.network_address}', '99', crc])
            command_str = header_str + ' ' + command_str

        return command_str.encode()

    def request_command(self, name: str, args: Iterable[Any]=[], flags: Dict[str, bool]={}, timeout: Optional[float]=None,
                        retries: Optional[int]=None, sequence_number: Optional[int]=None, reset_sequence: bool=False) -> str:
        if self.paused:
            while self.paused:
                time.sleep(0.5)

        if retries is None:
            retries = self.retries

        if sequence_number is None or reset_sequence:
            sequence_number = self.sequence_number
            self.sequence_number = (self.sequence_number + 1) % 98 + 1

        command = self.encode_command(name, flags, args, sequence_number)

        self.logger.info(f'Sending command: name={name}, args={args}, flags={flags}, retries={retries}')
        self.logger.debug(f'Sending request: {command}')
        self.stats.requests += 1
        self.stats.commands += 1

        try:
            crc_fail = False

            if not self.lock.acquire(timeout=self.lock_timeout):
                self.logger.critical(f'Controller {self.network_address} deadlocked')
                raise C9Error(C9Error.CONTROLLER_DEADLOCK, f'Controller {self.network_address} deadlocked')

            self.connection.write(command, timeout=self.write_timeout)
            response = self.connection.read_line(line_ending=b'\r\n', timeout=timeout or self.read_timeout)
            time.sleep(self.command_delay)
            self.lock.release()

            self.logger.debug(f'Response received: {response}')
            self.stats.responses += 1

            if not self.debug_protocol:
                try:
                    response, crc = response.rsplit(b'\r', 1)
                    response = response.strip(b'\r\x00')
                    response_crc = build_c9_crc(response + b'\r')
                    crc_fail = response_crc != crc.strip()
                except ValueError:
                    crc_fail = True

            try:
                # strip out any log statements
                lines = response.split(b'\r')
                logs = [line for line in lines if line.startswith(b'|')]
                response_lines = [line for line in lines if not line.startswith(b'|')]
                response = b'\n'.join(response_lines).strip().decode()

                self.logger.info(f'Response received: "{response}"')

                for log in logs:
                    self.device_logger.info(log.lstrip(b'|').decode())

            except UnicodeDecodeError:
                crc_fail = True

            if crc_fail or response.startswith('ERROR'):
                if crc_fail:
                    error_number = C9Error.INVALID_CRC
                    error_message = 'Invalid response CRC'
                else:
                    error_number, error_message = response.strip('ERROR ').split(':')

                error = C9Error(int(error_number), error_message.strip())

                self.stats.add_error(error)

                self.logger.warning(f'Response error: {error}')
                self.logger.debug(f'Errors: {self.stats.error_types_str}')

                if int(error_number) in self.RETRY_ERRORS and retries > 0:
                    return self.request_command(name, args, flags, timeout=timeout, retries=retries - 1, sequence_number=sequence_number, reset_sequence=reset_sequence)

                self.logger.error(f'Response error: [{error_number}] {error_message}')

                raise error

            return response

        except SerialException as error:
            self.lock.release()

            if retries == 0:
                self.logger.error(f'Serial error: {error}')
                raise error

            self.stats.add_error(error)

            self.logger.warning(f'Serial error: {error}')
            self.logger.debug(f'Errors: {self.stats.error_types_str}')

            timeout = retry_timeout_scale(self.retries, retries) * self.retry_timeout
            self.logger.debug(f'Retry timeout: {timeout}')
            time.sleep(timeout)

            return self.request_command(name, args, flags, timeout=timeout, retries=retries - 1, sequence_number=sequence_number, reset_sequence=reset_sequence)

    def register_output_handler(self, event, output, handler):
        if output not in self.output_handlers:
            self.output_handlers[output] = {}

        if event not in self.output_handlers[output]:
            self.output_handlers[output][event] = []

        self.output_handlers[output][event].append(handler)

    def dispatch_output_event(self, event, output: Optional[int]=None, *args, **kwargs):
        # if we weren't supplied an output, dispatch event to all handlers
        if output is None:
            for output, event_handlers in self.output_handlers.items():
                for handler in event_handlers.get(event, []):
                    handler(*args, **kwargs)
            return

        # otherwise look up the handler using the output and event
        if output not in self.output_handlers:
            return

        if event not in self.output_handlers[output]:
            return

        for handler in self.output_handlers[output][event]:
            handler(*args, **kwargs)

    def register_axis_handler(self, event, axis, handler):
        if axis not in self.axis_handlers:
            self.axis_handlers[axis] = {}

        if event not in self.axis_handlers[axis]:
            self.axis_handlers[axis][event] = []

        self.axis_handlers[axis][event].append(handler)

    def dispatch_axis_event(self, event, axis: Optional[int]=None, *args, **kwargs):
        # if we weren't supplied an axis, dispatch event to all handlers
        if axis is None:
            for axis, event_handlers in self.axis_handlers.items():
                for handler in event_handlers.get(event, []):
                    handler(*args, **kwargs)
            return

        # otherwise look up the handler using the axis and event
        if axis not in self.axis_handlers:
            return

        if event not in self.axis_handlers[axis]:
            return

        for handler in self.axis_handlers[axis][event]:
            handler(*args, **kwargs)

    def dispatch_axes_events(self, event, axes, *args, **kwargs):
        for axis in axes:
            self.dispatch_axis_event(event, axis, *args, **kwargs)

    def dispatch_position_events(self, axes, positions):
        axes_list = list(axes)
        for i in range(0, len(positions)):
            pass
            self.dispatch_axis_event('position', axes_list[i], int(positions[i]))

    def build_arm_args(self, x, y, z=None, gripper=None, velocity=None, acceleration=None):
        args = [x * 1000, y * 1000]

        if z is not None:
            args.append(z * 1000)

        if gripper is not None:
            args.append(gripper * 1000)

        if velocity is not None or acceleration is not None:
            if z is None:
                raise C9Error('ARMA command does not support velocity and acceleration without a z argument')
            if velocity is None or acceleration is None:
                raise C9Error('Both velocity and acceleration required')

            args += [velocity, acceleration]

        return args

    def build_flags(self, flags_dict: FlagsDict) -> List[str]:
        flags = []

        for flag, value in flags_dict.items():
            if isinstance(value, bool):
                if value:
                    flags.append(f'/{flag} 1')
            elif isinstance(value, float):
                flags.append(f'/{flag} {round(value)}')
            elif value is not None:
                flags.append(f'/{flag} {value}')

        return flags

    # Basic protocol
    def info(self) -> str:
        """
        Returns the firmware name and version of the connected C9 controller
        """
        return self.request_command('INFO')

    def ping(self, timeout: float=1.0, retries: int=5):
        """
        Sends a ping to the C9 controller, used to test connection
        """
        self.request_command('PING', timeout=timeout, retries=retries, reset_sequence=True)

    def address(self, new_address: int=None) -> int:
        """
        Return the address of the connected C9 controller, changing it to a new address if given

        :param new_address: optional new address to use for C9 controller
        :return: address of connected C9 controller
        """
        address = int(self.request_command('ADDR', [new_address]))
        if new_address is not None:
            self.network_address = new_address

        return address

    def home(self, *axes: int, if_needed: bool=True, skip: bool=False, timeout: int=240):
        """
        Homes the robot or axes connected to the C9 controller. If no axes are given, the main axes / robot will be homed.

        :param axes: axes to home, will home all main axes / robot if none given
        :param if_needed: only home if the connected C9 hasn't homed since power on
        :param skip: skip actual homing (WARNING: never use this)
        :param timeout: home timeout
        """
        response = self.request_command('HOME', axes or [None], flags={'C': if_needed, 'K': skip}, timeout=timeout)
        self.dispatch_axes_events('home', [0, 1, 2, 3])
        self.dispatch_position_events([0, 1, 2, 3], response.strip().split(' '))

    def elbow_bias(self, bias: Optional[int]=None) -> int:
        """
        Change the elbow bias of the N9

        :param bias: bias to use, one of: C9Controller.BIAS_MIN_SHOULDER, C9Controller.BIAS_MAX_SHOULDER, C9Controller.BIAS_CLOSEST
        :return: current elbow bias
        """
        args = [bias] if bias is not None else []
        return int(self.request_command('BIAS', args).strip())

    def swap_elbow(self):
        """
        Move the N9 to same effector position but with a different elbow position, changing elbow bias if needed
        """
        self.request_command('SWEL')
        self.wait_for_axes([1, 2])

    def axis_positions(self, *axes: int, units: bool=False, motor: bool=False) -> List[float]:
        """
        Return a list of current positions for the given axes

        :param axes: axes to find positions for
        :param units: use units instead of counts for positions if True
        :param motor: find actual position from motor controller instead of internal C9 value
        :return: list of axis positions
        """
        scale = 1000 if units else 1
        return [float(p)/scale for p in self.request_command('POS', list(axes), {'U': units, 'M': motor}).split(' ')]

    def axis_position(self, axis: int, units: bool=False, motor: bool=False) -> float:
        """
        Return the position of the given axis

        :param axis: axis number
        :param units: use units instead of counts for positions if True
        :param motor: find actual position from motor controller instead of internal C9 value
        :return: axis position
        """
        scale = 1000 if units else 1
        return float(self.request_command('POS', [axis], {'U': units, 'M': motor})) / scale

    def axis_velocity(self, axis: int) -> float:
        """
        Return the actual (not the target) velocity of the given axis

        :param axis: axis number
        :return: axis velocity in counts / s
        """
        return float(self.request_command('VEL', [axis]))

    def axis_current(self, axis: int, max_current: Optional[int]=None, max: bool=False) -> int:
        """
        Return the actual or max current for the given axis, setting the max current if given

        :param axis: axis number
        :param max_current: optional new max current value
        :param max: return max current instead of actual if True
        :return: actual or max current
        """
        if max_current is not None:
            max = True
        return int(self.request_command('CURR', [axis, max_current], flags={'M': max}))

    def cartesian_position(self) -> List[float]:
        """
        Return the x, y, z position of the N9

        :return: list containing x, y and z positions in mm
        """
        return [float(p)/1000 for p in self.request_command('POS', flags={'C': True}).split(' ')]

    def axes_moving(self, axes: List[int]) -> List[bool]:
        """
        Returns a list of moving statuses for the given axes (True if moving)

        :param axes: list of axis numbers
        :return: list of axis moving states
        """
        return [a == '1' for a in self.request_command('MVNG', axes).split(' ')]

    def axis_moving(self, axis: int) -> bool:
        """
        Returns given axis moving state

        :param axis: axis number
        :return: axis moving state (True if moving)
        """
        return self.request_command('MVNG', [axis]) == '1'

    def wait_for_axes(self, axes, poll_interval=0.01, verbose: bool=False):
        """
        Pauses until the given axes have stopped moving

        :param axes: list of axes numbers to wait for
        :param poll_interval: axis moving polling interval (defaults to 0.01 sec)
        :param verbose: controller verbosity while polling axes (defaults to False)
        """
        old_verbose = self.verbose
        if not verbose:
            self.verbose = False
        while True in self.axes_moving(axes):
            time.sleep(poll_interval)
        if not verbose:
            self.verbose = old_verbose

    def wait_for_axis(self, axis, poll_interval=0.01, verbose: bool=False):
        """
        Pause until the given axis has stopped moving

        :param axis: axis number
        :param poll_interval: axis moving polling interval (defaults to 0.01 sec)
        :param verbose: controller verbosity while polling axes (defaults to False)
        """
        self.wait_for_axes([axis], poll_interval, verbose)

    def wait_for_main_axes(self):
        """
        Pause until main axes / robot has stopped moving
        """
        self.wait_for_axes([0, 1, 2, 3])

    def output(self, output: int, state: Optional[bool]=None) -> bool:
        """
        Get or set the state of given output

        :param output: output number
        :param state: optional output state (True is on)
        """

        if state is None:
            return self.request_command('OUTP', [output]) == '1'

        value = 1 if state else 0
        self.request_command('OUTP', [output, value])
        self.dispatch_output_event('state', output, state)
        return state

    def output_toggle(self, output: int) -> bool:
        """
        Toggle the state of the given output

        :param output: output number
        """
        state = self.output(output)
        return self.output(output, not state)

    def outputs(self, pins: List[int]=[], all: bool=False) -> List[bool]:
        outputs = self.request_command('OUTP', pins, {'A': all}).split(' ')
        return [bool(int(value)) for value in outputs]

    def analog(self, pin: int) -> float:
        return float(self.request_command('ADC', [pin])) / 1000

    def analog_inputs(self, pins: List[int]=[], all: bool=False) -> List[float]:
        inputs = self.request_command('ADC', pins, {'A': all}).split(' ')
        return [float(value) / 1000 for value in inputs]

    def digital(self, pin: int) -> bool:
        return bool(int(self.request_command('INP', [pin])))

    def digital_inputs(self, pins: List[int]=[], all: bool=False) -> List[bool]:
        inputs = self.request_command('INP', pins, {'A': all}).split(' ')
        return [bool(int(value)) for value in inputs]

    def carousel(self, index: int, wait: bool=True, carousel: int=0):
        """
        Move attached carousels to the given index position (0 is the "home" position)

        :param index: carousel index
        :param wait: wait for carousel to stop (defaults to True)
        :param carousel: carousel index (defaults to 0)
        """
        self.request_command('CRSL', [carousel, index])

        if wait:
            self.wait_for_axis(self.AXIS_CAROUSEL)

    def weigh_scale(self, tare: bool=False) -> float:
        """
        Returns a measurement (in mg) from the attached weigh scale

        :param tare: tare the scale before measuring if True (default is False)
        :return: weight in mg
        """
        return float(self.request_command('WEIG', [], {'T': tare}))

    def elbow_length(self, length: Optional[float]=None) -> float:
        """
        Change the length of the last robot link, used for moving the robot probe

        :param length: length offset
        :return: current length offset
        """
        if length is None:
            return float(self.request_command('JLEN', []))
        else:
            return float(self.request_command('JLEN', [round(length * 1000)]))

    def use_probe(self, probe: bool=True, probe_offset: float=41.5):
        """
        Use the robot probe when moving

        :param probe: use probe if True (defaults to True)
        :param probe_offset: elbow length offset for probe (defaults to 41.5)
        """
        self.elbow_length(probe_offset if probe else 0)

    def speed(self, velocity: Optional[int]=None, acceleration: Optional[int]=None) -> Tuple[int, int]:
        """
        Get or sets the default velocity and acceleration for movements

        :param velocity: optional velocity in counts / s
        :param acceleration: optional acceleration in counts / s^2
        """
        velocity, acceleration = self.request_command('SPED', [velocity, acceleration]).split(' ')
        return int(velocity), int(acceleration)

    def move(self, axis_positions: Dict[int, float], velocity: Optional[int]=None, acceleration: Optional[int]=None,
             relative: bool=False, units: bool=False, wait: bool=True) -> List[float]:
        """
        Move axes to a relative or absolute position

        :param axis_positions: dictionary mapping axis numbers to target positions
        :param velocity: optional velocity during movement in counts / s
        :param acceleration: optional acceleration during movement in counts / s^2
        :param relative: use relative movements if True (defaults to False or absolute movements)
        :param units: positions are in units instead of counts (defaults to False)
        :param wait: wait for axes to stop moving (defaults to True)
        :return: list of final axis positions after move
        """
        if (velocity is None or acceleration is None) and velocity != acceleration:
            raise C9Error('Both acceleration and velocity need to be set or None')

        args = []

        for axis, position in axis_positions.items():
            args += [axis, position * 1000 if units else position]

        flags = {'V': velocity, 'A': acceleration, 'R': relative, 'U': units}
        response = self.request_command('MOVE', args, flags)
        start_time = time.time()
        duration, *positions = response.strip().split(' ')
        self.dispatch_position_events(axis_positions.keys(), positions)
        if wait:
            if self.move_prediction:
                duration = float(duration) * self.move_wait_scale / 1000 - (time.time() - start_time) - self.move_wait_delay
                time.sleep(max(duration, 0))
            self.wait_for_axes(axis_positions.keys())

        return positions

    def move_axis(self, axis: int, position: float, velocity: Optional[int]=None, acceleration: Optional[int]=None,
                  relative: bool=False, units: bool=False, wait: bool=True) -> float:
        """
        Moves an axis to an absolute or relative position

        :param axis: axis number
        :param position: target position
        :param velocity: optional velocity to use during move
        :param acceleration: optional acceleration to use during move
        :param relative: use relative movement if True (defaults to False)
        :param units: position is in units instead of counts if True (defaults to False)
        :param wait: wait for axis to stop moving if True (defaults to True)
        :return: final position after move
        """
        return self.move({axis: position}, velocity, acceleration, relative, units, wait)[0]

    def spin_axis(self, *axes: int, velocity: Optional[int]=None, acceleration: Optional[int]=None, stop: bool=False):
        """
        Start (or stop) spinning axes

        :param axes: axis numbers to spin
        :param velocity: velocity of spin
        :param acceleration: acceleration during spin
        :param stop: stops spin instead of starting if True (defaults to False)
        """
        self.request_command('SPIN', axes, {'V': velocity, 'A': acceleration, 'H': stop})

    def spin_axis_stop(self, *axes: int):
        """
        Stop spinning axes

        :param axes: axis numbers to stop
        """
        self.spin_axis(*axes, stop=True)

    def move_arm(self, x: Optional[float]=None, y: Optional[float]=None, z: Optional[float]=None,
                 gripper: Optional[float]=None, velocity: Optional[float]=None, acceleration: Optional[float]=None,
                 relative: bool=False, wait: bool=True) -> List[int]:
        """
        Move robot arm to a cartesian position, most positions are optional and the minimum number of axes will be moved

        :param x: optional x position in mm (x and y must be set as a pair)
        :param y: optional y position in mm (x and y must be set as a pair)
        :param z: optional z position in mm
        :param gripper: optional gripper position in degrees
        :param velocity: optional velocity during move in counts / s
        :param acceleration: optional acceleration during move in counts / s^2
        :param relative: perform a relative movement if True (defaults to False)
        :param wait: wait for movement to complete if True (defaults to False)
        :return: final positions of main axes after move
        """
        if (velocity is None or acceleration is None) and velocity != acceleration:
            raise C9Error('Both acceleration and velocity need to be set or None')

        axes = []
        flags = {'V': velocity, 'A': acceleration, 'R': relative}

        if x is not None and x is not False:
            flags['X'] = x * 1000
            axes += [1, 2]

        if y is not None and y is not False:
            flags['Y'] = y * 1000
            if 1 not in axes:
                axes += [1, 2]

        if z is not None and z is not False:
            flags['Z'] = z * 1000
            axes.append(3)

        if gripper is not None and gripper is not False:
            flags['G'] = gripper * 1000
            axes.append(0)

        response = self.request_command('ARM', [], flags)
        start_time = time.time()
        duration, *positions = response.strip().split(' ')
        self.dispatch_position_events([0, 1, 2, 3], positions)
        if wait:
            if self.move_prediction:
                duration = float(duration) * self.move_wait_scale
                time.sleep(max(float(duration) / 1000 - (time.time() - start_time) - self.move_wait_delay, 0))
            self.wait_for_axes(axes)

        return positions

    def com_init(self, port: int, baudrate: int):
        """
        Initialize an aux COM port on the C9 with the given baudrate

        :param port: COM port to initialize (matches COM number labelled on C9)
        :param baudrate: baudrate to to use (eg. 9600)
        """
        self.request_command('COM', [port, baudrate], {'I': True})

    def com_flush(self, port: int):
        """
        Flush the input buffer for the given aux COM port

        :param port: COM port to flush
        """
        self.request_command('COM', [port], {'F': True})

    def com_rx_size(self, port: int, retries: int=5) -> int:
        """
        Return the number of characters in the given aux COM port input buffer

        :param port: COM port to check
        :return: number of characters in input buffer
        """
        response = self.request_command('COM', [port], {'S': True})

        if not response:
            if retries <= 0:
                raise C9ResponseError(f'Invalid com_rx_size response: {response}')

            self.logger.warning(f'Invalid com_rx_size response: {response}')
            return self.com_rx_size(port, retries - 1)

        return int(response)

    def com_read(self, port: int, num_bytes: int=-1, timeout=1000) -> bytes:
        """
        Read bytes from the given aux COM port

        :param port: COM port to read from
        :param num_bytes: optional number of bytes to read (defaults to -1, which reads all bytes in input buffer)
        :return: bytes of data read from COM port
        """
        self.request_command('COM', [port, timeout, num_bytes], {'R': True})

        try:
            data = self.connection.read(num_bytes)
        except SerialReadTimeoutException:
            data = b''

        self.logger.debug(f'Received COM data: {data}')
        return data

    def com_write(self, port: int, data: bytes):
        """
        Write data to the given aux COM port

        :param port: COM port to write to
        :param data: data to write
        """

        self.request_command('COM', [port, len(data)], {'W': True})
        time.sleep(0.01)
        self.connection.write(data)
        self.logger.debug(f'Sent COM data: {data}')
        time.sleep(0.1)

    def com(self, port: int, baudrate: int=9600) -> Serial:
        """
        Create a new Serial class instance for the given aux COM port

        :param port: COM port to proxy serial data to
        :param baudrate: baudrate to use for COM port proxy
        :return: Serial instance
        """
        from north_c9.serial import C9SerialDevice
        return Serial(C9SerialDevice(self, port), baudrate=baudrate)

    def axis_state(self, axis: int) -> int:
        """
        Return the state of the given axis

        :param axis: axis to check
        :return: state of axis (-1 if in fault state)
        """
        return int(self.request_command('STAT', [axis]))

    def axis_error(self, axis: int) -> str:
        """
        Return the fault error code for the given axis

        :param axis: axis to check for errors
        :return: axis error code (0000 if no error)
        """
        return self.request_command('STAT', [axis], {'E': True})

    def axis_errors(self) -> Dict[int, int]:
        """
        Return a list of axis fault errors

        :return: list of (<axis number>, <fault code>) tuples, empty if no faults
        """
        raw_errors = self.request_command('STAT', [], {'F': True})
        axis_errors = raw_errors.split('\n')
        errors = {}

        for axis_error in axis_errors:
            axis, error = axis_error.split(' ')
            errors[int(axis)] = error.strip()

        return errors



from typing import List, Union, Callable, Optional, Dict
import time
from north_utils.joystick import Joystick
from north_c9.controller import C9Controller
from north_c9.joysticks import N9Joystick


class Keyframe:
    def __init__(self, x: Optional[float]=None, y: Optional[float]=None, z: Optional[float]=None, gripper: Optional[float]=None):
        self.x = x
        self.y = y
        self.z = z
        self.gripper = gripper

    def move_to(self, controller: C9Controller, **kwargs) -> None:
        controller.move_arm(x=self.x, y=self.y, z=self.z, gripper=self.gripper, **kwargs)


class Sequence:
    sequences: List['Sequence'] = []
    keyframes: List[Optional[Keyframe]] = []

    def __init__(self):
        self.current_index: int = 0
        self.playing = False
        self.paused = False

    @property
    def current_keyframe(self) -> Optional[Keyframe]:
        if len(self.keyframes) <= 0:
            return None

        return self.keyframes[self.current_index]

    def move_to_keyframe(self, controller: C9Controller, keyframe: Optional[Keyframe], **kwargs):
        if keyframe is None:
            return

        controller.move_arm(x=keyframe.x, y=keyframe.y, z=keyframe.z, gripper=keyframe.gripper, **kwargs)

    def move_to_current_keyframe(self, controller: C9Controller, **kwargs):
        self.move_to_keyframe(controller, keyframe=self.current_keyframe, **kwargs)

    def keyframe_step(self, controller: C9Controller, direction: int=1, **kwargs):
        self.current_index += direction

        if self.current_index >= len(self.keyframes):
            self.current_index = len(self.keyframes) - 1
            self.paused = True

        if self.current_index < 0:
            self.current_index = 0
            self.paused = True

        self.move_to_current_keyframe(controller, **kwargs)

    def move_to_next_keyframe(self, controller: C9Controller, **kwargs):
        self.keyframe_step(controller, direction=1, **kwargs)

    def move_to_previous_keyframe(self, controller: C9Controller, **kwargs):
        self.keyframe_step(controller, direction=-1, **kwargs)

    def insert_keyframe(self, keyframe: Keyframe):
        self.keyframes.insert(self.current_index, keyframe)
        self.current_index += 1

    def play(self, controller: C9Controller):
        self.playing = True
        self.move_to_current_keyframe(controller)
        while self.playing:
            if not self.paused:
                self.keyframe_step(controller)

            time.sleep(0.01)


class SequenceBuilder(N9Joystick):
    def __init__(self, controller: C9Controller, sequence: Optional[Sequence]=None):
        self.sequence = sequence if sequence is not None else Sequence()
        self.playback_mode = False
        N9Joystick.__init__(self, controller, moving=True)

    def start(self):
        N9Joystick.start(self)
        self.sequence.paused = True
        self.sequence.play(self.controller)

    def on_record_button(self, pressed: bool):
        if pressed and self.sequence.paused:
            x, y, z = self.controller.cartesian_position()
            gripper = self.controller.axis_position(self.GRIPPER_AXIS, units=True)
            keyframe = Keyframe(x=x, y=y, z=z, gripper=gripper)
            self.sequence.insert_keyframe(keyframe)

    def on_stop_button(self, pressed: bool):
        N9Joystick.on_stop_button(self, pressed)

        if pressed:
            self.sequence.paused = not self.sequence.paused

    def on_playback_mode_button(self, pressed: bool):
        self.playback_mode = pressed

    def on_keyframe_x_axis(self, value: float):
        if value < 0:
            self.sequence.move_to_previous_keyframe(self.controller)
        elif value > 0:
            self.sequence.move_to_next_keyframe(self.controller)

    def on_precise_x_axis_map(self, value: float):
        if not self.playback_mode:
            N9Joystick.on_x_precise_axis(self, value)
        else:
            self.on_keyframe_x_axis(value)

    def on_precise_y_axis_map(self, value: float):
        if not self.playback_mode:
            N9Joystick.on_y_precise_axis(self, value)

    mappings = {
        **N9Joystick.mappings,
        Joystick.RECORD_BUTTON: on_record_button,
        Joystick.STOP_BUTTON: on_stop_button,
        Joystick.PLAYBACK_MODE_BUTTON: on_playback_mode_button,
        Joystick.PRECISE_X_AXIS: on_precise_x_axis_map,
        Joystick.PRECISE_Y_AXIS: on_precise_y_axis_map,
    }



import unittest
from tests.util import *
from north_c9.util import Vector3
from north_c9.grid import Grid
from north_c9.deck import N9Deck, n9_deck


class GridTestSuite(unittest.TestCase):
    def test_deck(self):
        print("TEST DECK")
        assert_equal(N9Deck.A1, Vector3(375, -219))
        assert_equal(N9Deck.A2, Vector3(337.5, -219))
        assert_equal(N9Deck.A3, Vector3(300, -219))
        assert_equal(N9Deck.B1, Vector3(375, -181.5))
        assert_equal(N9Deck.B2, Vector3(337.5, -181.5))
        assert_equal(N9Deck.B3, Vector3(300, -181.5))
        assert_equal(N9Deck.Q21, Vector3(-375, 381))

        try:
            N9Deck.Q22
            raise AssertionError(f'N9Deck.Q22 is not unset')
        except AttributeError:
            pass

    def test_deck_grid(self):
        deck_grid = Grid(
            rows=21,
            columns=17,
            spacing=37.5,
            offset=Vector3(-375, -219)
        )
        assert_equal(deck_grid.locations['A1'], Vector3(375, -219))
        assert_equal(deck_grid.locations['A2'], Vector3(337.5, -219))
        assert_equal(deck_grid.locations['A3'], Vector3(300, -219))
        assert_equal(deck_grid.locations['B1'], Vector3(375, -181.5))
        assert_equal(deck_grid.locations['B2'], Vector3(337.5, -181.5))
        assert_equal(deck_grid.locations['B3'], Vector3(300, -181.5))

    def test_n9_grid_locations(self):
        for index, location in n9_deck.locations.items():
            assert_vector_equal(getattr(N9Deck, index), location)

    def test_grid(self):
        grid = Grid(2, 2, 10, Vector3(0, 0, 0))
        assert_equal(len(grid.grid), 2)
        assert_list_items_equal(grid.grid[0], [Vector3(10, 0, 0), Vector3(10, 10, 0)])
        assert_list_items_equal(grid.grid[1], [Vector3(0, 0, 0), Vector3(0, 10, 0)])

    def test_generate_class(self):
        grid = Grid(2, 2, 10, Vector3(0, 0, 0))
        source = grid.generate_class('GridTest', 'grid_test')
        assert_equal(source, """from north_c9.grid import Grid
from north_c9.util import Vector3


class GridTest(Grid):
    A1 = Vector3(x=10.0, y=0.0, z=0.0)
    B1 = Vector3(x=10.0, y=10.0, z=0.0)
    A2 = Vector3(x=0.0, y=0.0, z=0.0)
    B2 = Vector3(x=0.0, y=10.0, z=0.0)

    def __init__(self) -> None:
        Grid.__init__(self, rows=2, columns=2, spacing=10, offset=Vector3(x=0.0, y=0.0, z=0.0))


grid_test = GridTest()
""")


import unittest
from tests.util import *
from north_c9.kinematics import prrr_inverse_kinematics, prrr_forward_kinematics, PRRRBotConfiguration, PRRRBotDefinition
from north_c9.util import Vector3

zero_vector = Vector3(0, 0, 0)

bot_a = PRRRBotDefinition(column_offset=0, link1_length=5, link2_length=5, link3_length=0)
bot_b = PRRRBotDefinition(column_offset=0, link1_length=170.25, link2_length=170.25, link3_length=0)

zero_config = PRRRBotConfiguration(column=0, link1=0, link2=0, link3=0)
zero_config_position = Vector3(0, 10, 0)

simple_config = PRRRBotConfiguration(column=0, link1=math.pi/2, link2=math.pi/2, link3=0)
simple_config_position = Vector3(5, -5, 0)

class KinematicsTestSuite(unittest.TestCase):
    def test_forward_kinematics(self):
        position, rotation = prrr_forward_kinematics(bot_a, zero_config)
        assert_vector_equal(position, zero_config_position)

        position, rotation = prrr_forward_kinematics(bot_a, simple_config)
        assert_vector_equal(position, simple_config_position)

    def test_inverse_kinematics(self):
        config_a, config_b = prrr_inverse_kinematics(bot_a, zero_config_position, zero_vector, zero_vector)
        assert_config_equal(config_a, zero_config)
        assert_config_equal(config_b, zero_config)

        config_a, config_b = prrr_inverse_kinematics(bot_a, simple_config_position, Vector3(0, 0, math.pi), zero_vector)
        assert_config_equal(config_a, simple_config)

    def test_bot_b(self):
        position, rotation = prrr_forward_kinematics(bot_b, zero_config)
        assert_vector_equal(position, Vector3(0, 340.5, 0))

        config_a, config_b = prrr_inverse_kinematics(bot_b, Vector3(0, 340.5, 0), zero_vector, zero_vector)
        assert_config_equal(config_a, zero_config)

        config_a, config_b = prrr_inverse_kinematics(bot_b, Vector3(100, 100, 0), zero_vector, zero_vector)
        assert_config_equal(config_b, PRRRBotConfiguration(
            link1=1.9278841342163549, link2=-2.284971941637813, link3=-0.3570878074214583
        ))


import logging
import time
from ftdi_serial import Serial
from north_utils.location import Vector3
from north_utils.mixins import Runnable
from north_utils.joystick import Joystick, JoystickNotFoundError
from north_c9.controller import C9Controller, C9Error

logger = logging.getLogger(__name__)


class N9Joystick(Runnable):
    GRIPPER_AXIS = 0
    ELBOW_AXIS = 1
    SHOULDER_AXIS = 2
    COLUMN_AXIS = 3

    X_AXIS = 1
    Y_AXIS = 2
    Z_AXIS = 3

    GRIPPER_OUTPUT = 0
    PROBE_OUTPUT = 1

    def __init__(self, controller: C9Controller, velocity: int=40_000, acceleration: int=100_000,
                 moving: bool=False, ignore_missing_joystick: bool=False, z_offset: float=170.0):
        Runnable.__init__(self, logger)
        self.main_controller = controller
        self.controller = C9Controller(connection=controller.connection, use_joystick=False, ping=False, verbose=controller.verbose, read_timeout=0.2, debug_protocol=controller.debug_protocol)
        self.original_speed = (velocity, acceleration)
        self.origin_bias = C9Controller.BIAS_CLOSEST
        self.velocity = velocity
        self.acceleration = acceleration
        self.deadzone = 0.01
        self.moving = moving
        self.gripper_mode = False
        self.stop_timer_end = None
        self.stop_timer_timeout = 2.0
        self.base_velocity = 10.0
        self.x_velocity_scale = 0.0
        self.y_velocity_scale = 0.0
        self.z_velocity_scale = 0.0
        self.gripper_velocity_scale = 0.0
        self.speed_scale = 1.0
        self.precise_scale = 0.05
        self.slow_speed_scale = 0.5
        self.high_speed_scale = 5.0
        self.position = [0, 0, 0, 0]
        self.z_offset = z_offset
        self.recorded_position = None
        self.gripper_spinning = False

        try:
            self.joystick = Joystick()
        except JoystickNotFoundError as err:
            if ignore_missing_joystick:
                return

            raise err

        for event, handler in self.mappings.items():
            self.joystick.add_handler(event, self.wrap_handler(handler))

        self.start()

    def start(self):
        self.joystick.start()
        Runnable.start(self)

    def stop(self):
        Runnable.stop(self)
        self.joystick.stop()

    def stop_driving(self):
        pass

    def get_current_position(self):
        elbow = self.controller.axis_position(self.ELBOW_AXIS, units=True)
        shoulder = self.controller.axis_position(self.SHOULDER_AXIS, units=True)
        gripper = self.controller.axis_position(self.GRIPPER_AXIS, units=True)
        x, y, z = self.controller.cartesian_position()
        return [x, y, z, gripper + elbow + shoulder]

    def update_current_position(self):
        self.position = self.get_current_position()

    def start_moving(self):
        self.moving = True
        self.main_controller.paused = True
        self.stop_timer_end = None
        self.recorded_position = None

        self.original_speed = self.controller.speed()
        self.origin_bias = self.controller.elbow_bias()
        self.controller.speed(self.velocity, self.acceleration)
        self.controller.elbow_bias(C9Controller.BIAS_CLOSEST)
        # self.controller.home(if_needed=True)
        self.update_current_position()

    def stop_moving(self):
        self.moving = False
        self.main_controller.paused = False
        self.stop_timer_end = None
        self.controller.use_probe(False)
        self.controller.speed(*self.original_speed)
        self.controller.elbow_bias(self.origin_bias)

    def record_position(self, message: str, use_probe: bool=False) -> Vector3:
        print(message)
        self.start_moving()

        if use_probe:
            self.controller.use_probe(True)
            self.update_current_position()
        else:
            self.controller.use_probe(False)
            self.update_current_position()

        while self.recorded_position is None:
            time.sleep(0.01)

        return self.recorded_position

    def record_position_vector(self, message: str, use_probe: bool=False) -> Vector3:
        position = self.record_position(message, use_probe=use_probe)
        return Vector3(position[0], position[1], position[2])

    def record_z_offset(self, message: str='Drive the robot until the laser is focused to a point on the deck and press Record / Start', use_probe: bool=False):
        self.z_offset = self.record_z(message, offset=0, use_probe=use_probe)

    def record_z(self, message, offset=None, use_probe: bool=False):
        if offset is None:
            if self.z_offset is None:
                self.record_z_offset()

            offset = self.z_offset

        z_vector = self.record_position_vector(message, use_probe=use_probe)
        return z_vector.z() - offset

    def run(self):
        if self.moving:
            self.start_moving()

        while self.running:
            if self.stop_timer_end is not None:
                if self.moving:
                    if time.time() > self.stop_timer_end:
                        self.stop_moving()
                else:
                    self.start_moving()

            if self.moving:
                try:
                    position = [
                        self.position[0] + self.base_velocity * self.x_velocity_scale,
                        self.position[1] + self.base_velocity * self.y_velocity_scale,
                        self.position[2] + self.base_velocity * self.z_velocity_scale,
                        self.position[3] + self.base_velocity * self.gripper_velocity_scale,
                    ]

                    # normalize gripper position to -180.0 - 180.0
                    position[3] = (position[3] + 180) % 360 - 180

                    if self.x_velocity_scale != 0 or self.y_velocity_scale != 0 or self.z_velocity_scale != 0 or self.gripper_velocity_scale != 0:
                        self.controller.move_arm(*position)
                        self.position = position

                except C9Error as err:
                    if err.error_number not in (C9Error.INVALID_POSITION, C9Error.AXIS_MOVING):
                        raise err

    def wrap_handler(self, handler):
        return lambda state: handler(self, state)

    def axis_value(self, value: float) -> float:
        if abs(value) < self.deadzone:
            return 0

        return value

    def on_gripper_toggle_button(self, pressed: bool):
        if pressed:
            self.controller.output_toggle(self.GRIPPER_OUTPUT)

    def on_probe_toggle_button(self, pressed: bool):
        if pressed:
            self.controller.output_toggle(self.PROBE_OUTPUT)

    def on_elbow_swap_button(self, pressed: bool):
        if pressed:
            self.controller.swap_elbow()

    def on_spin_gripper_button(self, pressed: bool):
        if pressed:
            self.gripper_spinning = not self.gripper_spinning

    def on_gripper_mode_button(self, pressed: bool):
        self.gripper_mode = pressed

    def on_slow_speed_button(self, pressed: bool):
        if self.speed_scale == 1.0 or self.speed_scale == self.slow_speed_scale:
            self.speed_scale = self.slow_speed_scale if pressed else 1.0

    def on_high_speed_button(self, pressed: bool):
        if self.speed_scale == 1.0 or self.speed_scale == self.high_speed_scale:
            self.speed_scale = self.high_speed_scale if pressed else 1.0

    def on_stop_button(self, pressed: bool):
        if pressed:
            self.stop_timer_end = time.time() + self.stop_timer_timeout
        else:
            self.stop_timer_end = None

    def on_record_button(self, pressed: bool):
        if pressed:
            self.recorded_position = self.position

    def on_playback_mode_button(self, pressed: bool):
        pass

    def on_x_axis(self, value: float):
        self.x_velocity_scale = self.axis_value(value) * self.speed_scale * -1

    def on_y_axis(self, value: float):
        self.y_velocity_scale = self.axis_value(value) * self.speed_scale

    def on_z_axis(self, value: float):
        if not self.gripper_mode:
            self.z_velocity_scale = self.axis_value(value) * self.speed_scale * -1

    def on_gripper_left_axis(self, value: float):
        # trigger axis values range from -1 to 1, so add one to change range from 0 to 2
        self.gripper_velocity_scale = self.axis_value(value + 1) * self.speed_scale * -1

    def on_gripper_right_axis(self, value: float):
        self.gripper_velocity_scale = self.axis_value(value + 1) * self.speed_scale

    def on_x_precise_axis(self, value: float):
        self.x_velocity_scale = self.axis_value(value) * self.speed_scale * self.precise_scale * -1

    def on_y_precise_axis(self, value: float):
        self.y_velocity_scale = self.axis_value(value) * self.speed_scale * self.precise_scale * -1

    def on_gripper_precise_axis(self, value: float):
        self.gripper_velocity_scale = self.axis_value(value) * self.speed_scale * self.precise_scale

    def on_z_precise_axis(self, value: float):
        self.z_velocity_scale = self.axis_value(value) * self.speed_scale * self.precise_scale

    # mapping handlers
    def on_precise_x_axis_map(self, value: float):
        if self.gripper_mode:
            self.on_gripper_precise_axis(value)
            self.on_x_precise_axis(0.0)
        else:
            self.on_gripper_precise_axis(0.0)
            self.on_x_precise_axis(value)

    def on_precise_y_axis_map(self, value: float):
        if self.gripper_mode:
            self.on_y_precise_axis(0.0)
            self.on_z_precise_axis(value)
        else:
            self.on_y_precise_axis(value)
            self.on_z_precise_axis(0.0)

    mappings = {
        Joystick.GRIPPER_TOGGLE_BUTTON: on_gripper_toggle_button,
        Joystick.PROBE_TOGGLE_BUTTON: on_probe_toggle_button,
        Joystick.ELBOW_SWAP_BUTTON: on_elbow_swap_button,
        Joystick.SPIN_GRIPPER_BUTTON: on_spin_gripper_button,
        Joystick.STOP_BUTTON: on_stop_button,
        Joystick.RECORD_BUTTON: on_record_button,
        Joystick.SLOW_SPEED_BUTTON: on_slow_speed_button,
        Joystick.HIGH_SPEED_BUTTON: on_high_speed_button,
        Joystick.GRIPPER_MODE_BUTTON: on_gripper_mode_button,
        Joystick.PLAYBACK_MODE_BUTTON: on_playback_mode_button,

        Joystick.X_AXIS: on_x_axis,
        Joystick.Y_AXIS: on_y_axis,
        Joystick.GRIPPER_LEFT_AXIS: on_gripper_left_axis,
        Joystick.GRIPPER_RIGHT_AXIS: on_gripper_right_axis,
        Joystick.Z_AXIS: on_z_axis,
        Joystick.PRECISE_X_AXIS: on_precise_x_axis_map,
        Joystick.PRECISE_Y_AXIS: on_precise_y_axis_map
    }


import unittest
from tests.util import *
from north_c9.location import Frame, Location


class LocationTestSuite(unittest.TestCase):
    def test_rotation_x(self):
        assert_vector_equal(Location(1, 1, 0).rotate_x(90), Location(1, 0, 1))

    def test_rotation_y(self):
        assert_vector_equal(Location(1, 1, 0).rotate_y(90), Location(0, 1, -1))

    def test_rotation_z(self):
        assert_vector_equal(Location(1, 1, 0).rotate_z(90), Location(-1, 1, 0))

    def test_rotation(self):
        assert_vector_equal(Location(1, 1, 0).rotate(Location(90, 0, 0)), Location(1, 0, 1))
        assert_vector_equal(Location(1, 1, 0).rotate(Location(0, 90, 0)), Location(0, 1, -1))
        assert_vector_equal(Location(1, 1, 0).rotate(Location(0, 0, 90)), Location(-1, 1, 0))

    def test_transform(self):
        assert_vector_equal(Location(1, 1, 1).transform(Location(90, 0, 0), Location(0, 0, -1)), Location(1, -1, 0))
        assert_vector_equal(Location(1, 1, 1).transform(Location(0, 90, 0), Location(0, 0, -1)), Location(1, 1, -2))
        assert_vector_equal(Location(1, 1, 1).transform(Location(0, 0, 90), Location(0, 0, -1)), Location(-1, 1, 0))
        
    def test_frame_translation(self):
        frame = Frame(Location(1, 1, 0))
        assert_equal(frame.parent, Frame.WORLD)
        location = Location(1, 1, 1, frame=frame)
        assert_vector_equal(location.world_location, Location(2, 2, 1))

    def test_frame_rotation(self):
        frame = Frame(Location(0, 0, 0), Location(0, 0, 90))
        location = Location(1, 1, 0, frame=frame)
        assert_vector_equal(location.world_location, Location(-1, 1, 0))

    def test_frame_transform(self):
        frame = Frame(Location(1, 1, 0), Location(0, 0, 90))
        location = Location(1, 1, 0, frame=frame)
        assert_vector_equal(location.world_location, Location(0, 2, 0))

    def test_position_frame_transform(self):
        parent_frame = Frame(Location(1, 1, 0), Location(0, 0, 90))
        child_frame = Frame(Location(1, 1, 0), Location(0, 0, 90), parent_frame)
        location = Location(1, 1, 0, frame=child_frame)
        assert_vector_equal(location.world_location, Location(-1, 1, 0))



import time
from datetime import datetime
import sys
import random
from north_c9.controller import C9Controller, C9Error
from tests import power, spreadsheet

SPREADSHEET = '1TwUHwPSHzjxD5xRhM4n7-YtBNbcsS3RVntzqjVrOMcs'
SHEET = 'Home and Move'

VELOCITY = 5000
ACCELERATION = 20000

C9_SERIAL = 'A505PD7K'

campaign = []
cycles = 0
errors = 0

positions = [
    [150, 150, 50],
    [250, 200, 150],
    [-200, 50, 250],
    [-250, -200, 200],
]

last_position = None


def random_position():
    global last_position
    while True:
        position = random.choice(positions)
        if position != last_position:
            last_position = position
            return position


def random_move():
    while True:
        try:
            c9.move_arm(*random_position(), velocity=VELOCITY, acceleration=ACCELERATION)
            return
        except C9Error as err:
            if err.error_number != C9Error.INVALID_POSITION:
                raise err


def connect():
    for i in range(0, 10):
        try:
            return C9Controller(device_serial=C9_SERIAL)
        except:
            time.sleep(1)

    raise Exception("Error Connecting after 10 attempts!")


#spreadsheet.append_row(SPREADSHEET, SHEET, [datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f'), None, None])

#power.off()
#power.on()
c9 = connect()


while True:
    try:
        if cycles % 10 == 0:
            print()
            print(cycles, end=' ')

        c9.home(if_needed=False)

        for i in range(0, 2):
            random_move()

        cycles += 1
        print('.', end='')
        sys.stdout.flush()

    except Exception as err:
        errors += 1
        campaign.append(cycles)
        print()
        print('Error!', err)
        print('- cycles:', cycles)
        print('- errors:', errors)
        print('- campaign:', campaign)
        #spreadsheet.append_row(SPREADSHEET, SHEET, [datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f'), cycles, getattr(err, 'error_number', None), str(err)])
        cycles = 0
        c9.disconnect()
        power.off()
        power.on()

        c9 = connect()


import os
import unittest
import time
from ..util import *
from .. import mock_client
from north_c9.controller import C9Controller
from north_c9.robot import N9Robot, N9RobotMovementUnreachableLocation
from north_c9.util import Vector3, ZERO_VECTOR
from north_c9.grid import Grid
from north_c9.deck import n9_deck

mocked_client = os.environ.get('C9_MOCK_CLIENT', '') == 'True'
client = mock_client.C9Client() if mocked_client else None
controller = C9Controller(client=client)
robot = N9Robot(controller=controller, velocity_counts=5000, acceleration_counts=5000)


class DeckTestSuite(unittest.TestCase):
    def test_grid_locations(self):
        for index, location in n9_deck.locations.items():
            if not mocked_client:
                print(f'Moving to: {index} ({location})')

            try:
                location.z = 300
                robot.move_to_location(location, ZERO_VECTOR)
                time.sleep(1)
                #robot_location, robot_rotation = robot.forward_kinematics()
                #assert_vector_equal(robot_location, location, diff=0.00000000001)
            except N9RobotMovementUnreachableLocation:
                if not mocked_client:
                    print(' - unreachable')


import unittest
from tests.util import *
from north_c9 import trajectory


class TrajectoryTestSuite(unittest.TestCase):
    def test_trajectory_acceleration_triangular(self):
        traj = trajectory.Trajectory.acceleration_triangular(0, 10, 10)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_equal(len(traj.steps), 2)
        assert_list_items_equal(traj.steps[0].vap, [0, 10, 5])
        assert_list_items_equal(traj.steps[1].vap, [10, -10, 5])

        traj = trajectory.Trajectory.acceleration_triangular(10, 10, 10)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_list_items_equal(traj.steps[0].vap, [10, 0, 0])
        assert_list_items_equal(traj.steps[1].vap, [10, 0, 0])

        traj = trajectory.Trajectory.acceleration_triangular(10, 0, 10)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_list_items_equal(traj.steps[0].vap, [10, -10, 5])
        assert_list_items_equal(traj.steps[1].vap, [0, 10, 5])

        traj = trajectory.Trajectory.acceleration_triangular(10, 0, 10, direction=trajectory.Direction.REVERSE)
        assert_equal(traj.direction, trajectory.Direction.REVERSE)

    def test_trajectory_distance_triangular(self):
        traj = trajectory.Trajectory.distance_triangular(0, 10, 10)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_list_items_equal(traj.steps[0].vap, [0, 10, 5])
        assert_list_items_equal(traj.steps[1].vap, [10, -10, 5])

        traj = trajectory.Trajectory.distance_triangular(0, 10, -10)
        assert_equal(traj.direction, trajectory.Direction.REVERSE)
        assert_list_items_equal(traj.steps[0].vap, [0, 10, 5])
        assert_list_items_equal(traj.steps[1].vap, [10, -10, 5])

        traj = trajectory.Trajectory.distance_triangular(0, 10, 9)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_list_items_equal(traj.steps[0].vap, [0, 12, 4])
        assert_list_items_equal(traj.steps[1].vap, [10, -12, 5])

    def test_trajectory_trapezoidal(self):
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 100)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_equal(len(traj.steps), 3)
        assert_list_items_equal(traj.steps[0].vap, [0, 10, 5])
        assert_list_items_equal(traj.steps[1].vap, [10, 0, 90])
        assert_list_items_equal(traj.steps[2].vap, [10, -10, 5])

        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, -99)
        assert_equal(traj.direction, trajectory.Direction.REVERSE)
        assert_equal(len(traj.steps), 3)
        assert_list_items_equal(traj.steps[0].vap, [0, 10, 5])
        assert_list_items_equal(traj.steps[1].vap, [10, 0, 89])
        assert_list_items_equal(traj.steps[2].vap, [10, -10, 5])

        # should fall back to a triangular trajectory if the distance isn't long enough to ramp up fully
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 10)
        assert_equal(len(traj.steps), 2)
        assert_equal(traj.direction, trajectory.Direction.FORWARD)
        assert_list_items_equal(traj.steps[0].vap, [0, 10, 5])
        assert_list_items_equal(traj.steps[1].vap, [10, -10, 5])

    def test_trajectory_duration(self):
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 100)
        assert_equal(traj.duration, 11.0)

        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, -99)
        assert_equal(traj.duration, 10.9)

        # should fall back to a triangular trajectory if the distance isn't long enough to ramp up fully
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 10)
        assert_equal(traj.duration, 2.0)

    def test_trajectory_distance(self):
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 100)
        assert_equal(traj.distance, 100)

        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, -99)
        assert_equal(traj.distance, 99)

        # should fall back to a triangular trajectory if the distance isn't long enough to ramp up fully
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 10)
        assert_equal(traj.distance, 10)

    def test_trajectory_mopy(self):
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 100)
        assert_list_items_equal(traj.mopy, [1, 0, 10, 5, 10, 0, 90, 10, -10, 5])

        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, -99)
        assert_list_items_equal(traj.mopy, [0, 0, 10, 5, 10, 0, 89, 10, -10, 5])

        # should fall back to a triangular trajectory if the distance isn't long enough to ramp up fully
        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 10)
        assert_list_items_equal(traj.mopy, [1, 0, 10, 5, 10, -10, 5])

        traj = trajectory.Trajectory.trapezoidal(0, 10, 10, 9)
        assert_list_items_equal(traj.mopy, [1, 0, 12, 4, 10, -12, 5])


import unittest
import logging
import time
from ftdi_serial import Serial
from north_utils.test import assert_equal, assert_float_equal
from north_c9.controller import C9Controller
from north_c9.serial import C9SerialDevice
from north_devices.pumps.tecan_cavro import TecanCavro
from north_devices.scales.scientech_zeta import ScientechZeta

logging.basicConfig(level=logging.WARNING)
# logging.basicConfig(level=logging.DEBUG)

# C9_SERIAL = 'DA1FUC12'  # Propeller Quickstart board
# C9_SERIAL = 'A505PCDE'  # C9 Dev board
# C9_SERIAL = 'FT2FVF6U'  # C9 Network
C9_SERIAL = 'FT3FM1O1'
CLIENT_SERIAL = 'FTU9RIS7'
# CLIENT_SERIAL = 'FT2FVF6U'
COM_PORT = 1


class ComTestCase(unittest.TestCase):
    c9: C9Controller = None
    client: Serial = None

    @classmethod
    def setUpClass(cls):
        # cls.c9 = C9Controller(device_serial=C9_SERIAL, debug_protocol=True)
        cls.c9 = C9Controller(device_serial=C9_SERIAL)
        cls.client = Serial(device_serial=CLIENT_SERIAL)
        try:
            cls.c9.com_init(COM_PORT, 115200)
            time.sleep(1)
        except Exception as err:
            print(err)

    @classmethod
    def tearDownClass(cls):
        cls.c9.disconnect()
        cls.client.disconnect()

    def setUp(self):
        self.c9.com_init(COM_PORT, 115200)
        self.client.flush()
        self.c9.com_flush(COM_PORT)
        time.sleep(1)

    def test_flush(self):
        self.client.write(b'1234')
        size = self.c9.com_rx_size(COM_PORT)
        assert_equal(size, 4)
        self.c9.com_flush(COM_PORT)
        size = self.c9.com_rx_size(COM_PORT)
        assert_equal(size, 0)

    def test_size(self):
        message = b'test'
        self.client.write(message)
        time.sleep(0.1)
        size = self.c9.com_rx_size(COM_PORT)
        assert_equal(size, len(message))

    def test_read(self):
        message = b'foobar'
        self.client.write(message)
        time.sleep(0.1)
        data = self.c9.com_read(COM_PORT, len(message))
        assert_equal(data, message)

    def test_write(self):
        message = b'testing'
        self.c9.com_write(COM_PORT, message)
        time.sleep(0.1)
        data = self.client.read()
        assert_equal(data, message)

    def test_binary(self):
        self.client.set_parameters(baudrate=9600)
        self.c9.com_init(COM_PORT, 9600)

        message = b'\x0213k50R\x03?'
        self.c9.com_write(COM_PORT, message)
        time.sleep(0.1)
        data = self.client.read()
        assert_equal(data, message)
        time.sleep(0.1)
        message = b'\x0213k50R\x03?'
        self.client.write(message)
        time.sleep(0.1)
        data = self.c9.com_read(COM_PORT, len(message))
        assert_equal(data, message)

    def test_serial_device(self):
        device = C9SerialDevice(self.c9, COM_PORT)
        serial = Serial(device=device)

        message = b'test'
        serial.write(message)
        time.sleep(0.1)
        data = self.client.read()
        assert_equal(data, message)

        message = b'cool'
        self.client.write(message)
        time.sleep(0.1)
        data = serial.read()
        assert_equal(data, message)

    def test_request(self):
        serial = Serial(device=C9SerialDevice(self.c9, COM_PORT))

        request = b'request\r'
        response = b'response\r\n'
        self.client.write(response)
        time.sleep(0.1)
        response_data = serial.request(request, line_ending=b'\r\n')
        time.sleep(0.1)
        request_data = self.client.read()
        assert_equal(response_data, response.strip())
        assert_equal(request_data, request)

    def test_init(self):
        self.client.set_parameters(baudrate=9600)
        self.c9.com_init(COM_PORT, 9600)

        message = b'foobar'
        self.client.write(message)
        time.sleep(0.1)
        data = self.c9.com_read(COM_PORT, len(message))
        assert_equal(data, message)

        self.c9.com_init(COM_PORT, 115200)

    def test_cavro(self):
        serial = Serial(device=C9SerialDevice(self.c9, 0), baudrate=9600)
        cavro = TecanCavro(serial, 1)

        cavro.home()

        # Test dispense
        cavro.dispense_ml(0.1, 1, 2, velocity_ml=2)
        cavro.dispense_ml(0.5, 1, 2, velocity_counts=500)
        cavro.dispense_ml(2.5, 1, 2)

        # test velocity
        cavro.move_absolute_counts(1400, velocity_counts=500)
        cavro.move_absolute_counts(0)

        # test loop
        cavro.loop_start()
        cavro.move_absolute_counts(1000)
        cavro.move_absolute_counts(0)
        cavro.loop_end(2)
        cavro.execute()

    def test_cavro_network(self):
        serial = Serial(device=C9SerialDevice(self.c9, 0), baudrate=9600)
        cavro_a = TecanCavro(serial, 0)
        cavro_b = TecanCavro(serial, 1)

        TecanCavro.home_all()

        cavro_a.dispense_ml(0.5, 1, 2, wait=False)
        cavro_b.dispense_ml(0.5, 1, 2)

    def test_scale(self):
        serial = Serial(device=C9SerialDevice(self.c9, 1), baudrate=9600)
        scale = ScientechZeta(serial)
        scale.tare()
        weight = scale.weigh()
        assert_float_equal(weight, 0, diff=0.1)

    def test_cavro_and_scale(self):
        cavro = TecanCavro(self.c9.com(0), address=0)
        scale = ScientechZeta(self.c9.com(1))

        while True:
            cavro.home()
            scale.tare()
            scale.weigh()
            cavro.dispense_ml(0.1, 1, 2, velocity_ml=2)
            scale.tare()
            scale.weigh()
            cavro.dispense_ml(0.5, 1, 2, velocity_counts=500)
            scale.tare()
            scale.weigh()
            cavro.dispense_ml(2.5, 1, 2)


import unittest
from tests import mock_client
from tests.util import *
from north_c9 import robot, kinematics, controller
from north_c9.util import Vector3

home_config = kinematics.PRRRBotConfiguration(
    column=304.8,
    link1=-2.0465765369860796,
    link2=-2.5544228266482656,
    link3=0
)


def move_test(robot, location, rotation, config):
    robot.move_to_location(location, rotation)
    assert_config_equal(robot.configuration, config)


class RobotTestSuite(unittest.TestCase):
    def setUp(self):
        self.client = mock_client.C9Client()
        self.controller = controller.C9Controller(client=self.client)
        self.robot = robot.N9Robot(self.controller)

    def tearDown(self):
        self.client.disconnect()

    def test_init(self):
        assert_config_equal(self.robot.configuration, home_config)

    def test_move(self):
        move_test(self.robot, Vector3(100, 100, 304.8), Vector3(0, 0, 0), kinematics.PRRRBotConfiguration(
            column=304.8, link1=1.9278841342163549, link2=-2.284971941637813, link3=-0.3570878074214583
        ))

        move_test(self.robot, Vector3(-100, 200, 304.8), Vector3(0, 0, 0), kinematics.PRRRBotConfiguration(
            column=304.8, link1=0.39071237206491666, link2=-1.7087199621314455, link3=-1.318007590066529
        ))

        with self.controller.trace_trajectories() as trajectories:
            self.robot.move_to_location(Vector3(-100, 200, 304.8), Vector3(0, 0, 0))

            link1_distance = trajectories[2].distance
            link2_distance = trajectories[1].distance

            assert_float_equal(link1_distance, 26617)
            assert_float_equal(link2_distance, 6865)

            assert_list_items_equal(trajectories[2].mopy, [1, 0, 150000, 6750, 45000, 0, 13117, 45000, -150000, 6750])
            assert_list_items_equal(trajectories[1].mopy, [1, 0, 38682, 1741, 11605, 0, 3383, 11605, -38682, 1741])



from ftdi_serial import Device, SerialDeviceNotImplementedException
from north_c9.controller import C9Controller


class C9SerialDevice(Device):
    def __init__(self, controller: C9Controller, port: int):
        Device.__init__(self)
        self.controller = controller
        self.port = port
        self.read_timeout = 1000

    def clear(self):
        self.controller.com_flush(self.port)

    def set_baud_rate(self, baudrate: int):
        self.controller.com_init(self.port, baudrate)

    def set_timeouts(self, read_timeout: float, write_timeout: float):
        self.read_timeout = read_timeout

    def set_parameters(self, data_bits: int, stop_bits: int, parity_bits: int):
        if data_bits != 8 or stop_bits not in (0, 1, 2) or parity_bits != 0:
            raise SerialDeviceNotImplementedException(f'C9 serial devices only support 8 data bits, 1 stop bit and no parity')

    def get_input_size(self) -> int:
        return self.controller.com_rx_size(self.port)

    def get_output_size(self) -> int:
        raise SerialDeviceNotImplementedException(f'Cannot read output buffer size on C9 serial devices')

    def write(self, data: bytes):
        super().write(data)
        self.controller.com_write(self.port, data)

    def read(self, num_bytes: int, raw: bool = True) -> bytes:
        return self.controller.com_read(self.port, num_bytes)



from typing import Optional
import time
from north_utils.motion import degrees_to_counts, counts_to_degrees, mm_to_counts, counts_to_mm, cps_to_rpm, rpm_to_cps
from north_c9.controller import C9Controller


class AxisError(Exception):
    pass


class AxisMoveError(AxisError):
    pass


class JointError(Exception):
    pass


class RevoluteJointError(JointError):
    pass


class PrismaticJointError(JointError):
    pass


class Output:
    def __init__(self, controller: C9Controller, output_number: int):
        self.controller = controller
        self.output_number = output_number

    @property
    def state(self):
        return self.controller.output(self.output_number)

    def set_state(self, state):
        self.controller.output(self.output_number, state)

    def on(self):
        self.set_state(True)

    def off(self):
        self.set_state(False)

    def toggle(self):
        self.set_state(not self.state)


class OpenOutput(Output):
    def __init__(self, controller: C9Controller, output_number: int, open_state: bool=True):
        Output.__init__(self, controller, output_number)
        self.open_state = open_state

    def open(self, open_state: bool=True):
        state = self.open_state if open_state else not self.open_state
        self.set_state(state)

    def close(self):
        self.set_state(not self.open_state)


class Axis:
    def __init__(self,
                 controller: C9Controller,
                 axis_number: int,
                 velocity_counts: Optional[int]=None,
                 acceleration_counts: Optional[int]=None,
                 max_position: Optional[int]=None,
                 max_current: Optional[int]=None,
                 home_velocity_counts: Optional[int]=None,
                 home_acceleration_counts: Optional[int]=None,
                 home_max_current: Optional[int]=None,
                 home_to_block: bool=True,
                 main: bool=False,
                 reversed: bool=False,
                 **kwargs) -> None:
        self.controller = controller
        self.axis_number = axis_number
        self.velocity = velocity_counts
        self.acceleration = acceleration_counts
        self.max_current = max_current
        self.max_position = max_position
        self.reversed = reversed
        self.sign = -1 if reversed else 1
        self.main = main
        self.home_velocity = home_velocity_counts if home_velocity_counts is not None else velocity_counts or 0 / 2
        self.home_acceleration = home_acceleration_counts if home_acceleration_counts is not None else velocity_counts or 0 / 2
        self.home_max_current = home_max_current
        self.home_to_block = home_to_block and self.max_position is not None

        self.write_axis_settings()

    @property
    def position(self):
        return self.controller.axis_position(self.axis_number)

    @property
    def current_velocity(self):
        return self.controller.axis_velocity(self.axis_number)

    def write_axis_settings(self):
        if self.max_current is not None:
            self.set_max_current(self.max_current)

    def set_max_current(self, current, update: bool=True):
        self.controller.axis_current(self.axis_number, current, max=True)
        if update:
            self.max_current = current

    def home(self):
        if self.main:
            raise AxisError('Only aux axis can be homed manually')

        if self.home_max_current is not None:
            self.set_max_current(self.home_max_current)

        if self.home_to_block:
            self.move(-self.max_position * self.sign, velocity=self.home_velocity, acceleration=self.home_acceleration, relative=True, wait=True)

        self.controller.home(self.axis_number)

        if self.home_max_current is not None:
            self.set_max_current(self.max_current)

    def moving(self):
        return self.controller.axis_moving(self.axis_number)

    def wait(self):
        return self.controller.wait_for_axis(self.axis_number)

    def move(self, counts: int, velocity: Optional[int]=None, acceleration: Optional[int]=None, relative: bool=False,
             wait: bool=True):
        if velocity is None:
            velocity = self.velocity

        if acceleration is None:
            acceleration = self.acceleration

        if velocity is None and acceleration is None:
            raise AxisMoveError('Error moving joint, no velocity or acceleration given (and no defaults set for joint)')

        self.controller.move_axis(self.axis_number, counts, velocity, acceleration, wait=wait, units=self.main, relative=relative)


class RevoluteAxis(Axis):
    def __init__(self,
                 controller: C9Controller,
                 axis: int,
                 counts_per_revolution: float=1000.0,
                 zero_position_degrees: float=0.0,
                 position_degrees: float=0,
                 velocity_degrees: float=None,
                 acceleration_degrees: float=None,
                 inverted: bool=False,
                 **kwargs) -> None:
        kwargs['position'] = kwargs.get('position', degrees_to_counts(position_degrees, counts_per_revolution))
        kwargs['velocity'] = kwargs.get('velocity', degrees_to_counts(velocity_degrees, counts_per_revolution))
        kwargs['acceleration'] = kwargs.get('acceleration', degrees_to_counts(acceleration_degrees, counts_per_revolution))
        Axis.__init__(self, controller, axis, inverted=inverted, **kwargs)

        self.zero_position_degrees = zero_position_degrees
        self.counts_per_revolution = counts_per_revolution
        self.angle_sign = -1 if inverted else 1

    @property
    def position_degrees(self) -> float:
        return counts_to_degrees(self.position, self.counts_per_revolution, sign=self.angle_sign) - self.zero_position_degrees

    @property
    def velocity_degrees(self) -> float:
        return counts_to_degrees(self.velocity, self.counts_per_revolution) or 0

    @property
    def current_velocity_degrees(self) -> float:
        return counts_to_degrees(self.current_velocity, self.counts_per_revolution) or 0

    @property
    def current_velocity_rpm(self) -> float:
        return cps_to_rpm(self.current_velocity, self.counts_per_revolution) or 0

    @velocity_degrees.setter
    def velocity_degrees(self, velocity: float):
        self.velocity = degrees_to_counts(velocity, self.counts_per_revolution) or 0

    @property
    def acceleration_degrees(self) -> float:
        return counts_to_degrees(self.acceleration, self.counts_per_revolution) or 0

    @acceleration_degrees.setter
    def acceleration_degrees(self, acceleration: float):
        self.acceleration = degrees_to_counts(acceleration, self.counts_per_revolution)

    def move_degrees(self, degrees: float, velocity_degrees: Optional[float]=None,
                     acceleration_degrees: Optional[float]=None, relative: bool=False, wait: bool=True):

        if not relative:
            degrees += self.zero_position_degrees

        counts = degrees_to_counts(degrees * self.angle_sign, self.counts_per_revolution) or 0
        velocity = degrees_to_counts(velocity_degrees, self.counts_per_revolution)
        acceleration = degrees_to_counts(acceleration_degrees, self.counts_per_revolution)

        self.move(counts, velocity, acceleration, relative=relative, wait=wait)

    def spin(self, velocity_rpm: float, acceleration_rpm: float, duration: Optional[float]=None):
        velocity_counts = rpm_to_cps(velocity_rpm, self.counts_per_revolution)
        acceleration_counts = rpm_to_cps(acceleration_rpm, self.counts_per_revolution)

        self.controller.spin_axis(self.axis_number, velocity=velocity_counts, acceleration=acceleration_counts)

        if duration is not None:
            time.sleep(duration)
            self.spin_stop()

    def spin_stop(self, wait: bool=False):
        self.controller.spin_axis(self.axis_number, stop=True)

        if wait:
            self.controller.wait_for_axis(self.axis_number)


class PrismaticAxis(Axis):
    def __init__(self,
                 controller: C9Controller,
                 axis: int,
                 counts_per_mm: float=1.0,
                 position_mm: float=0.0,
                 zero_position_mm: float=0.0,
                 velocity_mm: Optional[float]=None,
                 acceleration_mm: Optional[float]=None,
                 inverted: bool=False,
                 **kwargs) -> None:
        kwargs['position'] = kwargs.get('position', mm_to_counts(position_mm, counts_per_mm))
        kwargs['velocity'] = kwargs.get('velocity', mm_to_counts(velocity_mm, counts_per_mm))
        kwargs['acceleration'] = kwargs.get('acceleration', mm_to_counts(acceleration_mm, counts_per_mm))
        Axis.__init__(self, controller, axis, **kwargs)

        self.counts_per_mm = counts_per_mm
        self.zero_position_mm = zero_position_mm
        self.sign = -1 if inverted else 1

    @property
    def position_mm(self) -> float:
        return counts_to_mm(self.position, self.counts_per_mm) * self.sign + self.zero_position_mm

    @property
    def velocity_mm(self) -> float:
        return counts_to_mm(self.velocity, self.counts_per_mm) or 0

    @velocity_mm.setter
    def velocity_mm(self, velocity: float):
        self.velocity = mm_to_counts(velocity, self.counts_per_mm) or 0

    @property
    def acceleration_mm(self) -> float:
        return counts_to_mm(self.acceleration, self.counts_per_mm) or 0

    @acceleration_mm.setter
    def acceleration_mm(self, acceleration: float):
        self.acceleration = mm_to_counts(acceleration, self.counts_per_mm) or 0

    def move_mm(self, mm: float, velocity_mm: Optional[float]=None, acceleration_mm: Optional[float]=None,
                relative: bool=False, wait: bool=True):
        if not relative:
            mm -= self.zero_position_mm

        counts = mm_to_counts(mm * self.sign, self.counts_per_mm) or 0
        velocity = mm_to_counts(velocity_mm, self.counts_per_mm)
        acceleration = mm_to_counts(acceleration_mm, self.counts_per_mm)

        self.move(counts, velocity, acceleration, relative=relative, wait=wait)



import time
from datetime import datetime
import sys
import random
from north_c9.controller import C9Controller, C9Error
from tests import power, spreadsheet, modbus

SPREADSHEET = '1TwUHwPSHzjxD5xRhM4n7-YtBNbcsS3RVntzqjVrOMcs'
SHEET = 'Move and Current'
MESSAGES_SHEET = 'Move and Current (Messages)'

C9_SERIAL = 'A505PD7K'
COM_SERIAL = 'FT2FT5C1'

VELOCITY = 5000
ACCELERATION = 20000

campaign = []
cycles = 0
errors = 0

positions = [
    [150, 150],
    [250, 200],
    [-200, 50],
    [-250, -200],
]

last_position = None

test_id = str(time.time())

#monitor = modbus.ModbusMonitor(test_id, device_serial=COM_SERIAL)
#monitor.start()

def random_position():
    global last_position
    z = random.randrange(100, 250)
    x_invert = random.choice([1, -1])
    while True:
        position = random.choice(positions)
        if position != last_position:
            last_position = position
            return [position[0] * x_invert, position[1], z]


def random_move():
    while True:
        try:
            c9.move_arm(*random_position(), velocity=VELOCITY, acceleration=ACCELERATION)
            return
        except C9Error as err:
            if err.error_number != C9Error.INVALID_POSITION:
                raise err


def find_modbus_timeout_axis():
    for i in range(0, 4):
        try:
            c9.axis_current(i)
        except C9Error:
            return i


def connect():
    for i in range(0, 10):
        try:
            return C9Controller(device_serial=C9_SERIAL)
        except:
            time.sleep(1)

    raise Exception("Error Connecting after 10 attempts!")


def save_cycle_data(recovered):
    axis = find_modbus_timeout_axis()
    row = [datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f'), test_id, axis, cycles, recovered, getattr(err, 'error_number', None), str(err)]
    spreadsheet.append_row(SPREADSHEET, SHEET, row)


#def save_message_data():
#    rows = [message.row() for message in monitor.get_messages()]
#    spreadsheet.append_rows(SPREADSHEET, MESSAGES_SHEET, rows)


#power.off()
#power.on()
c9 = connect()
c9.home()

while True:
    try:
        if cycles % 100 == 0:
            print()
            print(cycles, end=' ')

        #monitor.cycle = cycles
        random_move()
        for i in range(0, 4):
            c9.axis_current(i)

        cycles += 1
        print('.', end='')
        sys.stdout.flush()

    except Exception as err:
        errors += 1
        campaign.append(cycles)
        print()
        print('Error!', err)
        print('- cycles:', cycles)
        print('- errors:', errors)
        print('- campaign:', campaign)

        recovered = False

        if isinstance(err, C9Error) and err.error_number == C9Error.MODBUS_TIMEOUT:
            time.sleep(10)
            try:
                c9.axis_positions(0, 1, 2, 3, test=True)
                # if we've recovered from the timeout error, move on
                recovered = True
            except:
                pass

        save_cycle_data(recovered)
        #save_message_data()

        if not recovered:
            cycles = 0
            test_id = str(time.time())
            c9.disconnect()
            power.off()
            power.on()

            c9 = connect()
            c9.home()


import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = 'north_c9'
DESCRIPTION = 'A low-level communication library for the North C9 controller.'
URL = 'https://gitlab.com/north-robotics/north_c9'
EMAIL = 'sclark@northrobotics.com'
AUTHOR = 'Sean Clark'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    'ftdi_serial',
    'north_utils'
]

REQUIRED_REPOS = [
]

# What packages are optional?
EXTRAS = {
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests', 'docs', 'scripts')),
    install_requires=REQUIRED,
    dependency_links=REQUIRED_REPOS,
    extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
)


import logging
from north_c9.controller import C9Controller

logging.basicConfig(level=logging.WARNING)

c9 = C9Controller(device_serial='FT3FM1O1')
c9.home()

c9.speed(45000, 90000)

while True:
    c9.move_arm(150, 200, 200)
    c9.move_arm(-150, 150, 250)



import minimalmodbus

minimalmodbus.BAUDRATE = 38400
minimalmodbus.CLOSE_PORT_AFTER_EACH_CALL = False
minimalmodbus.STOPBITS = 2

invalid_responses = 0

devices = [
    minimalmodbus.Instrument('COM19', 1),
    #minimalmodbus.Instrument('COM19', 2),
    #minimalmodbus.Instrument('COM19', 3),
    #minimalmodbus.Instrument('COM19', 4),
]


def request(device, method, *args):
    global invalid_responses
    for i in range(0, 10):
        try:
            return getattr(device, method)(*args)

        # retry the message if we run into a recoverable error (up to 9 times)
        except Exception as err:
            if str(err).startswith('Checksum error') or \
                    str(err).startswith('Too short') or \
                    str(err).startswith('No communication') or \
                    str(err).startswith('Wrong return slave'):
                if i >= 9:
                    raise err

                invalid_responses += 1
                print('Invalid responses: ', invalid_responses)
                # flush the input buffers so minimalmodbus doesn't go out of sync
                device.serial.reset_input_buffer()
            else:
                raise err


def home():
    for device in devices:
        # start homing (change state to 2)
        request(device, 'write_long', 6002, 2)

    # wait until the state of all devices is no longer 2
    while True:
        homing = False
        for device in devices:
            if request(device, 'read_long', 5000) == 2:
                homing = True

        if not homing:
            return


# attempt to home in a loop until we run into an error (after retrying the message 9 times)
while True:
    home()



import unittest
from tests.util import *
from tests import mock_client


class ClientTestSuite(unittest.TestCase):
    def setUp(self):
        self.client = mock_client.C9Client(verbose=True)

    def tearDown(self):
        self.client.disconnect()

    def test_echo(self):
        response = self.client.echo()
        assert_equal(response, b'<ECHOSUCKA>\r')

    def test_home(self):
        self.client.connection.write(b'a')
        response = self.client.home()
        assert_equal(response, b'<HOME>\r')